# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical.ext.resource import ResourceType
from .create import Create
from ..params import ResourceRepository


class DefaultCreate(Create[ResourceType]):
    creates = True
    detail = False
    exists = False
    method = 'POST'
    requires_body = True
    status_code = 201
    verb = 'create'

    async def handle(
        self,
        repo: ResourceRepository,
        obj: ResourceType
    ) -> ResourceType:
        return await obj.persist(repo)